@@ -1119,6 +1119,7 @@ post_install_foreign(Pack, PackDir, Options) :-
         ),
         build_steps(BuildSteps, PackDir, Options)
     ).
+post_install_foreign(_, _, _).
 
 
 %!  foreign_present(+PackDir) is semidet.
@@ -2106,7 +2107,7 @@ message(install_downloaded(File)) -->
       size_file(File, Size) },
     [ 'Install "~w" (~D bytes)'-[Base, Size] ].
 message(git_post_install(PackDir, Pack)) -->
-    (   { is_foreign_pack(PackDir) }
+    (   { is_foreign_pack(PackDir, _) }
     ->  [ 'Run post installation scripts for pack "~w"'-[Pack] ]
     ;   [ 'Activate pack "~w"'-[Pack] ]
     ).
