@@ -3,7 +3,7 @@
     Author:        Vitor Santos Costa
     E-mail:        vscosta@gmail.com
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2007-2013, Vitor Santos Costa
+    Copyright (c)  2007-2021, Vitor Santos Costa
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -33,38 +33,36 @@
 */
 
 :- module(rbtrees,
-          [rb_new/1,
-           rb_empty/1,                  % ?T
-           rb_lookup/3,                 % +Key, -Value, +T
-           rb_update/4,                 % +T, +Key, +NewVal, -TN
-           rb_update/5,                 % +T, +Key, ?OldVal, +NewVal, -TN
-           rb_apply/4,                  % +T, +Key, :G, -TN
-           rb_lookupall/3,              % +Key, -Value, +T
-           rb_insert/4,                 % +T0, +Key, ?Value, -TN
-           rb_insert_new/4,             % +T0, +Key, ?Value, -TN
-           rb_delete/3,                 % +T, +Key, -TN
-           rb_delete/4,                 % +T, +Key, -Val, -TN
-           rb_visit/2,                  % +T, -Pairs
-           rb_visit/3,
-           rb_keys/2,                   % +T, +Keys
-           rb_keys/3,
-           rb_map/2,
-           rb_map/3,
-           rb_partial_map/4,
-           rb_clone/3,
-           rb_clone/4,
-           rb_min/3,
-           rb_max/3,
-           rb_del_min/4,
-           rb_del_max/4,
-           rb_next/4,
-           rb_previous/4,
-           list_to_rbtree/2,
-           ord_list_to_rbtree/2,
-           is_rbtree/1,
-           rb_size/2,
-           rb_in/3
-       ]).
+          [ rb_new/1,                   % -Tree
+            rb_empty/1,                 % ?Tree
+            rb_lookup/3,                % +Key, -Value, +T
+            rb_update/4,                % +Tree, +Key, +NewVal, -NewTree
+            rb_update/5,                % +Tree, +Key, ?OldVal, +NewVal, -NewTree
+            rb_apply/4,                 % +Tree, +Key, :G, -NewTree
+            rb_insert/4,                % +T0, +Key, ?Value, -NewTree
+            rb_insert_new/4,            % +T0, +Key, ?Value, -NewTree
+            rb_delete/3,                % +Tree, +Key, -NewTree
+            rb_delete/4,                % +Tree, +Key, -Val, -NewTree
+            rb_visit/2,                 % +Tree, -Pairs
+            rb_keys/2,                  % +Tree, +Keys
+            rb_map/2,                   % +Tree, :Goal
+            rb_map/3,                   % +Tree, :Goal, -MappedTree
+            rb_partial_map/4,           % +Tree, +Keys, :Goal, -MappedTree
+            rb_fold/4,                  % :Goal, +Tree, +State0, -State
+            rb_clone/3,                 % +TreeIn, -TreeOut, -Pairs
+            rb_min/3,                   % +Tree, -Key, -Value
+            rb_max/3,                   % +Tree, -Key, -Value
+            rb_del_min/4,               % +Tree, -Key, -Val, -TreeDel
+            rb_del_max/4,               % +Tree, -Key, -Val, -TreeDel
+            rb_next/4,                  % +Tree, +Key, -Next, -Value
+            rb_previous/4,              % +Tree, +Key, -Next, -Value
+            list_to_rbtree/2,           % +Pairs, -Tree
+            ord_list_to_rbtree/2,       % +Pairs, -Tree
+            is_rbtree/1,                % @Tree
+            rb_size/2,                  % +Tree, -Size
+            rb_in/3                     % ?Key, ?Value, +Tree
+          ]).
+:- autoload(library(error), [domain_error/2]).
 
 /** <module> Red black trees
 
@@ -79,15 +77,29 @@ Nil-node, a node shared for each nil-node in  the tree. Any node has the
 form colour(Left, Key, Value, Right), where _colour_  is one of =red= or
 =black=.
 
-@author Vitor Santos Costa, Jan Wielemaker
+@author Vitor Santos Costa, Jan Wielemaker, Samer Abdallah,
+        Peter Ludemann.
 @see "Introduction to Algorithms", Second Edition Cormen, Leiserson,
      Rivest, and Stein, MIT Press
 */
 
+% rbtrees.pl is derived from YAP's rbtrees.yap, with some minor editing.
+% One difference is that the SWI-Prolog version  assumes that a key only
+% appears once in the tree - the   YAP  code is somewhat inconsistent in
+% that  (and  even  allows  rb_lookup/3  to    backtrack,  plus  it  has
+% rb_lookupall/3, which isn't in the SWI-Prolog code).
+
+% The code has also been modified to   use SWI-Prolog's '=>' operator to
+% throw an existence_error(matching_rule, _)  exception   if  Tree isn't
+% instantiated (if ':-' is used, an  uninstanted   Tree  gets  set to an
+% empty tree, which probably isn't the desired result).
+
 :- meta_predicate
-    rb_map(+,:,-),
-    rb_partial_map(+,+,:,-),
-    rb_apply(+,+,:,-).
+    rb_map(+,2,-),
+    rb_map(?,1),
+    rb_partial_map(+,+,2,-),
+    rb_apply(+,+,2,-),
+    rb_fold(3,+,+,-).
 
 /*
 :- use_module(library(type_check)).
@@ -112,28 +124,31 @@ form colour(Left, Key, Value, Right), where _colour_  is one of =red= or
 :- pred next(tree(K,V),K,pair(K,V),V,tree(K,V)).
 */
 
-%!  rb_new(-T) is det.
+%!  rb_new(-Tree) is det.
 %
-%   Create a new Red-Black tree.
+%   Create a new Red-Black tree Tree.
 %
 %   @deprecated     Use rb_empty/1.
 
+:- det(rb_new/1).
 rb_new(t(Nil,Nil)) :-
     Nil = black('',_,_,'').
 
-%!  rb_empty(?T) is semidet.
+%!  rb_empty(?Tree) is semidet.
 %
-%   Succeeds if T is an empty Red-Black tree.
+%   Succeeds if Tree is an empty Red-Black tree.
 
 rb_empty(t(Nil,Nil)) :-
     Nil = black('',_,_,'').
 
-%!  rb_lookup(+Key, -Value, +T) is semidet.
+%!  rb_lookup(+Key, -Value, +Tree) is semidet.
 %
-%   Backtrack through all elements with  key   Key  in the Red-Black
-%   tree T, returning for each the value Value.
+%   True when Value is associated with Key   in the Red-Black tree Tree.
+%   The given Key may include variables, in   which  case the RB tree is
+%   searched for a key with equivalent,   as  in (==)/2, variables. Time
+%   complexity is O(log N) in the number of elements in the tree.
 
-rb_lookup(Key, Val, t(_,Tree)) :-
+rb_lookup(Key, Val, t(_,Tree)) =>
     lookup(Key, Val, Tree).
 
 lookup(_, _, black('',_,_,'')) :- !, fail.
@@ -151,11 +166,11 @@ lookup(<, K, V, Tree) :-
 lookup(=, _, V, Tree) :-
     arg(3,Tree,V).
 
-%!  rb_min(+T, -Key, -Value) is semidet.
+%!  rb_min(+Tree, -Key, -Value) is semidet.
 %
-%   Key is the minimum key in T, and is associated with Val.
+%   Key is the minimum key in Tree, and is associated with Val.
 
-rb_min(t(_,Tree), Key, Val) :-
+rb_min(t(_,Tree), Key, Val) =>
     min(Tree, Key, Val).
 
 min(red(black('',_,_,_),Key,Val,_), Key, Val) :- !.
@@ -165,11 +180,11 @@ min(red(Right,_,_,_), Key, Val) :-
 min(black(Right,_,_,_), Key, Val) :-
     min(Right,Key,Val).
 
-%!  rb_max(+T, -Key, -Value) is semidet.
+%!  rb_max(+Tree, -Key, -Value) is semidet.
 %
-%   Key is the maximal key in T, and is associated with Val.
+%   Key is the maximal key in Tree, and is associated with Val.
 
-rb_max(t(_,Tree), Key, Val) :-
+rb_max(t(_,Tree), Key, Val) =>
     max(Tree, Key, Val).
 
 max(red(_,Key,Val,black('',_,_,_)), Key, Val) :- !.
@@ -179,12 +194,12 @@ max(red(_,_,_,Left), Key, Val) :-
 max(black(_,_,_,Left), Key, Val) :-
     max(Left,Key,Val).
 
-%!  rb_next(+T, +Key, -Next,-Value) is semidet.
+%!  rb_next(+Tree, +Key, -Next, -Value) is semidet.
 %
-%   Next is the next element after Key  in T, and is associated with
-%   Val.
+%   Next is the next element after Key   in Tree, and is associated with
+%   Val. Fails if Key isn't in Tree or if Key is the maximum key.
 
-rb_next(t(_,Tree), Key, Next, Val) :-
+rb_next(t(_,Tree), Key, Next, Val) =>
     next(Tree, Key, Next, Val, []).
 
 next(black('',_,_,''), _, _, _, _) :- !, fail.
@@ -207,12 +222,13 @@ next(=, _, _, _, NK, Val, Tree, Candidate) :-
     ;   Candidate = (NK-Val)
     ).
 
-%!  rb_previous(+T, +Key, -Previous, -Value) is semidet.
+%!  rb_previous(+Tree, +Key, -Previous, -Value) is semidet.
 %
-%   Previous is the  previous  element  after   Key  in  T,  and  is
-%   associated with Val.
+%   Previous  is  the  previous  element  after  Key  in  Tree,  and  is
+%   associated with Val. Fails if Key isn't  in   Tree  or if Key is the
+%   minimum key.
 
-rb_previous(t(_,Tree), Key, Previous, Val) :-
+rb_previous(t(_,Tree), Key, Previous, Val) =>
     previous(Tree, Key, Previous, Val, []).
 
 previous(black('',_,_,''), _, _, _, _) :- !, fail.
@@ -235,16 +251,18 @@ previous(=, _, _, _, K, Val, Tree, Candidate) :-
     ;   Candidate = (K-Val)
     ).
 
-%!  rb_update(+T, +Key, +NewVal, -TN) is semidet.
-%!  rb_update(+T, +Key, ?OldVal, +NewVal, -TN) is semidet.
+%!  rb_update(+Tree, +Key, +NewVal, -NewTree) is semidet.
+%!  rb_update(+Tree, +Key, ?OldVal, +NewVal, -NewTree) is semidet.
 %
-%   Tree TN is tree T,  but  with   value  for  Key  associated with
-%   NewVal.  Fails if it cannot find Key in T.
+%   Tree NewTree is tree Tree, but with   value  for Key associated with
+%   NewVal. Fails if it cannot find Key in Tree.
 
-rb_update(t(Nil,OldTree), Key, OldVal, Val, t(Nil,NewTree)) :-
+rb_update(t(Nil,OldTree), Key, OldVal, Val, NewTree2) =>
+    NewTree2 = t(Nil,NewTree),
     update(OldTree, Key, OldVal, Val, NewTree).
 
-rb_update(t(Nil,OldTree), Key, Val, t(Nil,NewTree)) :-
+rb_update(t(Nil,OldTree), Key, Val, NewTree2) =>
+    NewTree2 = t(Nil,NewTree),
     update(OldTree, Key, _, Val, NewTree).
 
 update(black(Left,Key0,Val0,Right), Key, OldVal, Val, NewTree) :-
@@ -271,14 +289,15 @@ update(red(Left,Key0,Val0,Right), Key, OldVal, Val, NewTree) :-
         update(Right, Key, OldVal, Val, NewRight)
     ).
 
-%!  rb_apply(+T, +Key, :G, -TN) is semidet.
+%!  rb_apply(+Tree, +Key, :G, -NewTree) is semidet.
 %
-%   If the value associated with  key  Key   is  Val0  in  T, and if
-%   call(G,Val0,ValF) holds, then TN differs from T only in that Key
-%   is associated with value ValF in  tree   TN.  Fails if it cannot
-%   find Key in T, or if call(G,Val0,ValF) is not satisfiable.
+%   If the value associated  with  key  Key   is  Val0  in  Tree, and if
+%   call(G,Val0,ValF) holds, then NewTree differs from Tree only in that
+%   Key is associated with value  ValF  in   tree  NewTree.  Fails if it
+%   cannot find Key in Tree, or if call(G,Val0,ValF) is not satisfiable.
 
-rb_apply(t(Nil,OldTree), Key, Goal, t(Nil,NewTree)) :-
+rb_apply(t(Nil,OldTree), Key, Goal, NewTree2) =>
+    NewTree2 = t(Nil,NewTree),
     apply(OldTree, Key, Goal, NewTree).
 
 %apply(black('',_,_,''), _, _, _) :- !, fail.
@@ -314,17 +333,18 @@ apply(red(Left,Key0,Val0,Right), Key, Goal,
         apply(Right, Key, Goal, NewRight)
     ).
 
-%!  rb_in(?Key, ?Val, +Tree) is nondet.
+%!  rb_in(?Key, ?Value, +Tree) is nondet.
 %
-%   True if Key-Val appear in Tree. Uses indexing if Key is bound.
+%   True when Key-Value is a key-value pair in red-black tree Tree. Same
+%   as below, but does not materialize the pairs.
+%
+%        rb_visit(Tree, Pairs), member(Key-Value, Pairs)
+%
+%   Leaves a choicepoint, even  if  Key   is  instantiated;  to  avoid a
+%   choicepoint, use rb_lookup/3.
 
-rb_in(Key, Val, t(_,T)) :-
-    var(Key),
-    !,
+rb_in(Key, Val, t(_,T)) =>
     enum(Key, Val, T).
-rb_in(Key, Val, t(_,T)) :-
-    lookup(Key, Val, T).
-
 
 enum(Key, Val, black(L,K,V,R)) :-
     L \= '',
@@ -339,32 +359,6 @@ enum_cases(Key, Val, _, _, _, R) :-
     enum(Key, Val, R).
 
 
-%!  rb_lookupall(+Key, -Value, +T)
-%
-%   Lookup all elements with  key  Key   in  the  red-black  tree T,
-%   returning the value Value.
-
-rb_lookupall(Key, Val, t(_,Tree)) :-
-    lookupall(Key, Val, Tree).
-
-
-lookupall(_, _, black('',_,_,'')) :- !, fail.
-lookupall(Key, Val, Tree) :-
-    arg(2,Tree,KA),
-    compare(Cmp,KA,Key),
-    lookupall(Cmp,Key,Val,Tree).
-
-lookupall(>, K, V, Tree) :-
-    arg(4,Tree,NTree),
-    rb_lookupall(K, V, NTree).
-lookupall(=, _, V, Tree) :-
-    arg(3,Tree,V).
-lookupall(=, K, V, Tree) :-
-    arg(1,Tree,NTree),
-    lookupall(K, V, NTree).
-lookupall(<, K, V, Tree) :-
-    arg(1,Tree,NTree),
-    lookupall(K, V, NTree).
 
                  /*******************************
                  *       TREE INSERTION         *
@@ -372,13 +366,15 @@ lookupall(<, K, V, Tree) :-
 
 % We don't use parent nodes, so we may have to fix the root.
 
-%!  rb_insert(+T0, +Key, ?Value, -TN) is det.
+%!  rb_insert(+Tree, +Key, ?Value, -NewTree) is det.
 %
-%   Add an element with key Key and Value  to the tree T0 creating a
-%   new red-black tree TN. If Key  is   a  key in T0, the associated
-%   value is replaced by Value.  See also rb_insert_new/4.
+%   Add an element with key Key and Value   to  the tree Tree creating a
+%   new red-black tree NewTree. If Key is  a key in Tree, the associated
+%   value is replaced by Value. See also rb_insert_new/4.
 
-rb_insert(t(Nil,Tree0),Key,Val,t(Nil,Tree)) :-
+:- det(rb_insert/4).
+rb_insert(t(Nil,Tree0),Key,Val,NewTree) =>
+    NewTree = t(Nil,Tree),
     insert(Tree0,Key,Val,Nil,Tree).
 
 
@@ -436,12 +432,13 @@ insert2(black(L,K0,V0,R), K, V, Nil, NT, Flag) :-
 
 % We don't use parent nodes, so we may have to fix the root.
 
-%!  rb_insert_new(+T0, +Key, ?Value, -TN) is semidet.
+%!  rb_insert_new(+Tree, +Key, ?Value, -NewTree) is semidet.
 %
-%   Add a new element with key Key and Value  to the tree T0 creating a
-%   new red-black tree TN.   Fails if Key is a key in T0.
+%   Add a new element with key Key and Value to the tree Tree creating a
+%   new red-black tree NewTree. Fails if Key is a key in Tree.
 
-rb_insert_new(t(Nil,Tree0),Key,Val,t(Nil,Tree)) :-
+rb_insert_new(t(Nil,Tree0),Key,Val,NewTree) =>
+    NewTree = t(Nil,Tree),
     insert_new(Tree0,Key,Val,Nil,Tree).
 
 insert_new(Tree0,Key,Val,Nil,Tree) :-
@@ -524,39 +521,42 @@ fix_right(not_done,Tmp,Final,Done) :-
 % case 1 of RB: just need to change colors.
 %
 fix_right(black(red(Ep,KD,VD,De),KC,VC,red(red(Ga,KB,VB,Be),KA,VA,Al)),
-        red(black(Ep,KD,VD,De),KC,VC,black(red(Ga,KB,VB,Be),KA,VA,Al)),
-        not_done) :- !.
+          red(black(Ep,KD,VD,De),KC,VC,black(red(Ga,KB,VB,Be),KA,VA,Al)),
+          not_done) :- !.
 fix_right(black(red(Ep,KD,VD,De),KC,VC,red(Ga,Ka,Va,red(Be,KB,VB,Al))),
-        red(black(Ep,KD,VD,De),KC,VC,black(Ga,Ka,Va,red(Be,KB,VB,Al))),
-        not_done) :- !.
+          red(black(Ep,KD,VD,De),KC,VC,black(Ga,Ka,Va,red(Be,KB,VB,Al))),
+          not_done) :- !.
 %
 % case 2 of RB: got a knee so need to do rotations
 %
 fix_right(black(De,KC,VC,red(red(Ga,KB,VB,Be),KA,VA,Al)),
-        black(red(De,KC,VC,Ga),KB,VB,red(Be,KA,VA,Al)),
-        done) :- !.
+          black(red(De,KC,VC,Ga),KB,VB,red(Be,KA,VA,Al)),
+          done) :- !.
 %
 % case 3 of RB: got a line
 %
 fix_right(black(De,KC,VC,red(Ga,KB,VB,red(Be,KA,VA,Al))),
-        black(red(De,KC,VC,Ga),KB,VB,red(Be,KA,VA,Al)),
-        done) :- !.
+          black(red(De,KC,VC,Ga),KB,VB,red(Be,KA,VA,Al)),
+          done) :- !.
 %
 % case 4 of RB: nothing to do.
 %
 fix_right(T,T,done).
 
 
-%!  rb_delete(+T, +Key, -TN).
-%!  rb_delete(+T, +Key, -Val, -TN).
+%!  rb_delete(+Tree, +Key, -NewTree).
+%!  rb_delete(+Tree, +Key, -Val, -NewTree).
 %
-%   Delete element with key Key from the tree T, returning the value
-%   Val associated with the key and a new tree TN.
+%   Delete element with key Key from the  tree Tree, returning the value
+%   Val associated with the key and a new tree NewTree.
+%   Fails if Key is not in Tree.
 
-rb_delete(t(Nil,T), K, t(Nil,NT)) :-
+rb_delete(t(Nil,T), K, NewTree) =>
+    NewTree = t(Nil,NT),
     delete(T, K, _, NT, _).
 
-rb_delete(t(Nil,T), K, V, t(Nil,NT)) :-
+rb_delete(t(Nil,T), K, V, NewTree) =>
+    NewTree = t(Nil,NT),
     delete(T, K, V0, NT, _),
     V = V0.
 
@@ -590,12 +590,14 @@ delete(black(L,_,V,R), _, V, OUT, Flag) :-
     % K == K0,
     delete_black_node(L,R,OUT,Flag).
 
-%!  rb_del_min(+T, -Key, -Val, -TN)
+%!  rb_del_min(+Tree, -Key, -Val, -NewTree)
 %
-%   Delete the least element from the tree T, returning the key Key,
-%   the value Val associated with the key and a new tree TN.
+%   Delete the least element from the tree  Tree, returning the key Key,
+%   the value Val associated with the key and a new tree NewTree.
+%   Fails if Tree is empty.
 
-rb_del_min(t(Nil,T), K, Val, t(Nil,NT)) :-
+rb_del_min(t(Nil,T), K, Val, NewTree) =>
+    NewTree = t(Nil,NT),
     del_min(T, K, Val, Nil, NT, _).
 
 del_min(red(black('',_,_,_),K,V,R), K, V, Nil, OUT, Flag) :-
@@ -612,12 +614,15 @@ del_min(black(L,K0,V0,R), K, V, Nil, NT, Flag) :-
     fixup_left(Flag0,black(NL,K0,V0,R),NT, Flag).
 
 
-%!  rb_del_max(+T, -Key, -Val, -TN)
+%!  rb_del_max(+Tree, -Key, -Val, -NewTree)
 %
-%   Delete the largest element from the   tree  T, returning the key
-%   Key, the value Val associated with the key and a new tree TN.
+%   Delete the largest element from  the   tree  Tree, returning the key
+%   Key, the value Val associated with the key and a new tree NewTree.
+%   Fails if Tree is empty.
 
-rb_del_max(t(Nil,T), K, Val, t(Nil,NT)) :-
+
+rb_del_max(t(Nil,T), K, Val, NewTree) =>
+    NewTree = t(Nil,NT),
     del_max(T, K, Val, Nil, NT, _).
 
 del_max(red(L,K,V,black('',_,_,_)), K, V, Nil, OUT, Flag) :-
@@ -777,38 +782,31 @@ fixup3(black(black(red(Fi,KE,VE,Ep),KD,VD,C),KB,VB,black(Be,KA,VA,Al)),
        black(black(Fi,KE,VE,Ep),KD,VD,black(C,KB,VB,black(Be,KA,VA,Al))),
        done).
 
-
-%
-% whole list
+%!  rb_visit(+Tree, -Pairs) is det.
 %
+%   Pairs is an infix visit of tree Tree, where each element of Pairs is
+%   of the form Key-Value.
 
-%!  rb_visit(+T, -Pairs)
-%
-%   Pairs is an infix visit of tree   T, where each element of Pairs
-%   is of the form K-Val.
-
-rb_visit(t(_,T),Lf) :-
+:- det(rb_visit/2).
+rb_visit(t(_,T),Lf) =>
     visit(T,[],Lf).
 
-rb_visit(t(_,T),L0,Lf) :-
-    visit(T,L0,Lf).
-
-visit(black('',_,_,_),L,L) :- !.
-visit(red(L,K,V,R),L0,Lf) :-
+visit(black('',_,_,_),L0,L) => L0 = L.
+visit(red(L,K,V,R),L0,Lf) =>
     visit(L,[K-V|L1],Lf),
     visit(R,L0,L1).
-visit(black(L,K,V,R),L0,Lf) :-
+visit(black(L,K,V,R),L0,Lf) =>
     visit(L,[K-V|L1],Lf),
     visit(R,L0,L1).
 
-:- meta_predicate rb_map(?,:,?). % this is not strictly required
-:- meta_predicate map(?,:,?,?).  % this is required.
+:- meta_predicate map(?,2,?,?).  % this is required.
 
 %!  rb_map(+T, :Goal) is semidet.
 %
 %   True if call(Goal, Value) is true for all nodes in T.
 
-rb_map(t(Nil,Tree),Goal,t(Nil,NewTree)) :-
+rb_map(t(Nil,Tree),Goal,NewTree2) =>
+    NewTree2 = t(Nil,NewTree),
     map(Tree,Goal,NewTree,Nil).
 
 
@@ -824,17 +822,16 @@ map(black(L,K,V,R),Goal,black(NL,K,NV,NR),Nil) :-
     map(L,Goal,NL,Nil),
     map(R,Goal,NR,Nil).
 
-:- meta_predicate rb_map(?,:). % this is not strictly required
-:- meta_predicate map(?,:).  % this is required.
+:- meta_predicate map(?,1).  % this is required.
 
-%!  rb_map(+T, :G, -TN) is semidet.
+%!  rb_map(+Tree, :G, -NewTree) is semidet.
 %
-%   For all nodes Key in the tree   T,  if the value associated with
-%   key Key is Val0 in tree T,  and if call(G,Val0,ValF) holds, then
-%   the  value  associated  with  Key  in   TN  is  ValF.  Fails  if
-%   call(G,Val0,ValF) is not satisfiable for all Var0.
+%   For all nodes Key in the tree Tree, if the value associated with key
+%   Key is Val0 in tree Tree, and   if call(G,Val0,ValF) holds, then the
+%   value  associated  with  Key  in   NewTree    is   ValF.   Fails  if
+%   call(G,Val0,ValF) is not satisfiable for all Val0.
 
-rb_map(t(_,Tree),Goal) :-
+rb_map(t(_,Tree),Goal) =>
     map(Tree,Goal).
 
 
@@ -850,13 +847,41 @@ map(black(L,_,V,R),Goal) :-
     map(L,Goal),
     map(R,Goal).
 
-%!  rb_clone(+T, -NT, -Pairs)
+%!  rb_fold(:Goal, +Tree, +State0, -State).
 %
-%   "Clone" the red-back tree into a new  tree with the same keys as
-%   the original but with all values set to unbound values. Nodes is
-%   a list containing all new nodes as pairs K-V.
+%   Fold the given predicate  over  all   the  key-value  pairs in Tree,
+%   starting with initial state State0  and   returning  the final state
+%   State. Pred is called as
+%
+%       call(Pred, Key-Value, State1, State2)
+%
+%   Determinism depends on Goal.
 
-rb_clone(t(Nil,T),t(Nil,NT),Ns) :-
+rb_fold(Pred, t(_,T), S1, S2) =>
+    fold(T, Pred, S1, S2).
+
+fold(black(L,K,V,R), Pred) -->
+    (   {L == ''}
+    ->  []
+    ;   fold_parts(Pred, L, K-V, R)
+    ).
+fold(red(L,K,V,R), Pred) -->
+    fold_parts(Pred, L, K-V, R).
+
+fold_parts(Pred, L, KV, R) -->
+    fold(L, Pred),
+    call(Pred, KV),
+    fold(R, Pred).
+
+%!  rb_clone(+TreeIn, -TreeOut, -Pairs) is det.
+%
+%   `Clone' the red-back tree TreeIn into a   new  tree TreeOut with the
+%   same keys as the original but with all values set to unbound values.
+%   Pairs is a list containing all new nodes as pairs K-V.
+
+:- det(rb_clone/3).
+rb_clone(t(Nil,T),TreeOut,Ns) =>
+    TreeOut = t(Nil,NT),
     clone(T,Nil,NT,Ns,[]).
 
 clone(black('',_,_,''),Nil,Nil,Ns,Ns) :- !.
@@ -867,26 +892,17 @@ clone(black(L,K,_,R),Nil,black(NL,K,NV,NR),NsF,Ns0) :-
     clone(L,Nil,NL,NsF,[K-NV|Ns1]),
     clone(R,Nil,NR,Ns1,Ns0).
 
-rb_clone(t(Nil,T),ONs,t(Nil,NT),Ns) :-
-    clone(T,Nil,ONs,[],NT,Ns,[]).
-
-clone(black('',_,_,''),Nil,ONs,ONs,Nil,Ns,Ns) :- !.
-clone(red(L,K,V,R),Nil,ONsF,ONs0,red(NL,K,NV,NR),NsF,Ns0) :-
-    clone(L,Nil,ONsF,[K-V|ONs1],NL,NsF,[K-NV|Ns1]),
-    clone(R,Nil,ONs1,ONs0,NR,Ns1,Ns0).
-clone(black(L,K,V,R),Nil,ONsF,ONs0,black(NL,K,NV,NR),NsF,Ns0) :-
-    clone(L,Nil,ONsF,[K-V|ONs1],NL,NsF,[K-NV|Ns1]),
-    clone(R,Nil,ONs1,ONs0,NR,Ns1,Ns0).
-
-%!  rb_partial_map(+T, +Keys, :G, -TN)
+%!  rb_partial_map(+Tree, +Keys, :G, -NewTree)
 %
-%   For all nodes Key in Keys, if  the value associated with key Key
-%   is Val0 in tree T,  and   if  call(G,Val0,ValF)  holds, then the
-%   value associated with Key  in  TN  is   ValF.  Fails  if  or  if
-%   call(G,Val0,ValF) is not satisfiable for  all Var0. Assumes keys
-%   are not repeated.
+%   For all nodes Key in Keys, if the   value associated with key Key is
+%   Val0 in tree Tree, and if   call(G,Val0,ValF)  holds, then the value
+%   associated with Key in NewTree is ValF,   otherwise  it is the value
+%   associated with the key in Tree. Fails if   Key  isn't in Tree or if
+%   call(G,Val0,ValF) is not satisfiable for all   Val0 in Keys. Assumes
+%   keys are sorted and not repeated (fails if this is not true).
 
-rb_partial_map(t(Nil,T0), Map, Goal, t(Nil,TF)) :-
+rb_partial_map(t(Nil,T0), Map, Goal, NewTree) =>
+    NewTree = t(Nil,TF),
     partial_map(T0, Map, [], Nil, Goal, TF).
 
 partial_map(T,[],[],_,_,T) :- !.
@@ -925,20 +941,15 @@ partial_map(black(L,K,V,R),Map,MapF,Nil,Goal,black(NL,K,NV,NR)) :-
     ).
 
 
+%!  rb_keys(+Tree, -Keys) is det.
 %
-% whole keys
-%
-%%      rb_keys(+T, -Keys)
-%
-%       Keys is unified with  an  ordered  list   of  all  keys  in  the
-%       Red-Black tree T.
+%   Keys is unified with an ordered list   of  all keys in the Red-Black
+%   tree Tree.
 
-rb_keys(t(_,T),Lf) :-
+:- det(rb_keys/2).
+rb_keys(t(_,T),Lf) =>
     keys(T,[],Lf).
 
-rb_keys(t(_,T),L0,Lf) :-
-    keys(T,L0,Lf).
-
 keys(black('',_,_,''),L,L) :- !.
 keys(red(L,K,_,R),L0,Lf) :-
     keys(L,[K|L1],Lf),
@@ -948,26 +959,35 @@ keys(black(L,K,_,R),L0,Lf) :-
     keys(R,L0,L1).
 
 
-%!  list_to_rbtree(+L, -T) is det.
+%!  list_to_rbtree(+List, -Tree) is det.
 %
-%   T is the red-black tree corresponding to the mapping in list L.
+%   Tree is the red-black tree  corresponding   to  the mapping in List,
+%   which should be a list of Key-Value   pairs. List should not contain
+%   more than one entry for each distinct key, but this is not validated
+%   by list_to_rbtree/2.
 
+:- det(list_to_rbtree/2).
 list_to_rbtree(List, T) :-
     sort(List,Sorted),
     ord_list_to_rbtree(Sorted, T).
 
-%!  ord_list_to_rbtree(+L, -T) is det.
+%!  ord_list_to_rbtree(+List, -Tree) is det.
 %
-%   T is the red-black tree corresponding  to the mapping in ordered
-%   list L.
+%   Tree is the red-black tree  corresponding   to  the  mapping in list
+%   List, which should be a list  of   Key-Value  pairs. List should not
+%   contain more than one entry for each   distinct key, but this is not
+%   validated by ord_list_to_rbtree/2. List is assumed
+%   to be sorted according to the standard order of terms.
 
-ord_list_to_rbtree([], t(Nil,Nil)) :-
-    !,
+:- det(ord_list_to_rbtree/2).
+ord_list_to_rbtree([], Tree) =>
+    Tree = t(Nil,Nil),
     Nil = black('', _, _, '').
-ord_list_to_rbtree([K-V], t(Nil,black(Nil,K,V,Nil))) :-
-    !,
+ord_list_to_rbtree([K-V], Tree), nonvar(K) =>
+    Tree = t(Nil,black(Nil,K,V,Nil)),
     Nil = black('', _, _, '').
-ord_list_to_rbtree(List, t(Nil,Tree)) :-
+ord_list_to_rbtree(List, Tree2) =>
+    Tree2 = t(Nil,Tree),
     Nil = black('', _, _, ''),
     Ar =.. [seq|List],
     functor(Ar,_,L),
@@ -993,11 +1013,12 @@ build_node( 0, Left, K, Val, Right, red(Left, K, Val, Right)) :- !.
 build_node( _, Left, K, Val, Right, black(Left, K, Val, Right)).
 
 
-%!  rb_size(+T, -Size) is det.
+%!  rb_size(+Tree, -Size) is det.
 %
-%   Size is the number of elements in T.
+%   Size is the number of elements in Tree.
 
-rb_size(t(_,T),Size) :-
+:- det(rb_size/2).
+rb_size(t(_,T),Size) =>
     size(T,0,Size).
 
 size(black('',_,_,_),Sz,Sz) :- !.
@@ -1010,60 +1031,83 @@ size(black(L,_,_,R),Sz0,Szf) :-
     size(L,Sz1,Sz2),
     size(R,Sz2,Szf).
 
-%!  is_rbtree(?Term) is semidet.
+%!  is_rbtree(@Term) is semidet.
 %
-%   True if Term is a valide Red-Black tree.
-%
-%   @tbd    Catch variables.
+%   True if Term is a valid Red-Black   tree. Processes the entire tree,
+%   checking the coloring of the nodes, the  balance and the ordering of
+%   keys. It does _not_ validate that keys are sufficiently instantiated
+%   to ensure the tree remains valid if a key is further instantiated.
 
-is_rbtree(X) :-
-    var(X), !, fail.
-is_rbtree(t(Nil,Nil)) :- !.
-is_rbtree(t(_,T)) :-
-    catch(rbtree1(T), msg(_,_), fail).
+is_rbtree(X), var(X) =>
+    fail.
+is_rbtree(t(Nil,Nil)) => true.
+is_rbtree(t(_,T)) =>
+    Err = error(_,_),
+    catch(check_rbtree(T), Err, is_rbtree_error(Err)).
+
+is_rbtree_error(Err), Err = error(resource_error(_),_) => throw(Err).
+is_rbtree_error(_) => fail.
 
 %
 % This code checks if a tree is ordered and a rbtree
 %
+% TODO: Use (?=)/2 to verify that pairwise keys are strictly
+%       ordered, no matter how the keys become instantiated.
+%       This is not a complete test; to be completely safe, all
+%       use of compare/3 (and (@<)/2 etc) would need to also
+%       use (?=)/2, which would be expensive.
 
-rbtree1(black(L,K,_,R)) :-
+check_rbtree(black(L,K,_,R)) =>
     find_path_blacks(L, 0, Bls),
     check_rbtree(L,-inf,K,Bls),
     check_rbtree(R,K,+inf,Bls).
-rbtree1(red(_,_,_,_)) :-
-    throw(msg("root should be black",[])).
+check_rbtree(Node), Node = red(_,_,_,_) =>
+    domain_error(rb_black, Node).
 
 
-find_path_blacks(black('',_,_,''), Bls, Bls) :- !.
-find_path_blacks(black(L,_,_,_), Bls0, Bls) :-
+find_path_blacks(black('',_,_,''), Bls0, Bls) => Bls = Bls0.
+find_path_blacks(black(L,_,_,_), Bls0, Bls) =>
     Bls1 is Bls0+1,
     find_path_blacks(L, Bls1, Bls).
-find_path_blacks(red(L,_,_,_), Bls0, Bls) :-
+find_path_blacks(red(L,_,_,_), Bls0, Bls) =>
     find_path_blacks(L, Bls0, Bls).
 
-check_rbtree(black('',_,_,''),Min,Max,Bls0) :-
-    !,
+check_rbtree(black('',_,_,''),Min,Max,Bls0) =>
     check_height(Bls0,Min,Max).
-check_rbtree(red(L,K,_,R),Min,Max,Bls) :-
+check_rbtree(red(L,K,_,R),Min,Max,Bls) =>
     check_val(K,Min,Max),
     check_red_child(L),
     check_red_child(R),
     check_rbtree(L,Min,K,Bls),
     check_rbtree(R,K,Max,Bls).
-check_rbtree(black(L,K,_,R),Min,Max,Bls0) :-
+check_rbtree(black(L,K,_,R),Min,Max,Bls0) =>
     check_val(K,Min,Max),
     Bls is Bls0-1,
     check_rbtree(L,Min,K,Bls),
     check_rbtree(R,K,Max,Bls).
 
-check_height(0,_,_) :- !.
-check_height(Bls0,Min,Max) :-
-    throw(msg("Unbalance ~d between ~w and ~w~n",[Bls0,Min,Max])).
+check_height(0,_,_) => true.
+check_height(Bls0,Min,Max) =>
+    throw(error(rbtree(balance(Bls0, Min, Max)), _)).
 
-check_val(K, Min, Max) :- ( K @> Min ; Min == -inf), (K @< Max ; Max == +inf), !.
-check_val(K, Min, Max) :-
-    throw(msg("not ordered: ~w not between ~w and ~w~n",[K,Min,Max])).
+check_val(K, Min, Max), (K @> Min ; Min == -inf), (K @< Max ; Max == +inf) =>
+    true.
+check_val(K, Min, Max) =>
+    throw(error(rbtree(order(K, Min, Max)), _)).
 
-check_red_child(black(_,_,_,_)).
-check_red_child(red(_,K,_,_)) :-
-    throw(msg("must be red: ~w~n",[K])).
+check_red_child(black(_,_,_,_)) => true.
+check_red_child(Node), Node = red(_,_,_,_) =>
+    domain_error(rb_black, Node).
+
+
+		 /*******************************
+		 *            MESSAGES		*
+		 *******************************/
+
+:- multifile
+    prolog:error_message//1.
+
+prolog:error_message(rbtree(balance(Bls0, Min, Max))) -->
+    [ 'Unbalance ~d between ~w and ~w'-[Bls0,Min,Max] ].
+prolog:error_message(rbtree(order(K, Min, Max))) -->
+    [ 'not ordered: ~w not between ~w and ~w'-[K,Min,Max] ].
