@@ -36,6 +36,7 @@
 :- module(pengines_io,
           [ pengine_writeln/1,          % +Term
             pengine_nl/0,
+            pengine_tab/1,
             pengine_flush_output/0,
             pengine_format/1,           % +Format
             pengine_format/2,           % +Format, +Args
@@ -160,6 +161,20 @@ pengine_nl :-
 pengine_nl :-
     nl.
 
+%!  pengine_tab(+N)
+%
+%   Emit N spaces
+
+pengine_tab(N) :-
+    pengine_output,
+    !,
+    length(List, N),
+    maplist(=(&(nbsp)), List),
+    send_html(List).
+pengine_tab(N) :-
+    tab(N).
+
+
 %!  pengine_flush_output
 %
 %   No-op.  Pengines do not use output buffering (maybe they should
@@ -569,11 +584,12 @@ binding_residual('Residual'  = '$residual'(Residual),   'Residual', [Residual])
 binding_residual_clauses(
     '_wfs_residual_program' = '$wfs_residual_program'(Delays, Clauses),
     '_wfs_residual_program', Residuals, Clauses) :-
-    phrase(comma_list(Delays), Residuals).
+    phrase(delay_list(Delays), Residuals).
 
-comma_list(true) --> !.
-comma_list((A,B)) --> !, comma_list(A), comma_list(B).
-comma_list(A) --> [A].
+delay_list(true) --> !.
+delay_list((A,B)) --> !, delay_list(A), delay_list(B).
+delay_list(M:A) --> !, [M:'$wfs_undefined'(A)].
+delay_list(A) --> ['$wfs_undefined'(A)].
 
 add_projection(-, _, JSON, JSON) :- !.
 add_projection(VarNames0, ResVars0, JSON0, JSON) :-
@@ -642,6 +658,12 @@ term_html(Term, Vars, WriteOptions) -->
     { nonvar(Term) },
     binding_term(Term, Vars, WriteOptions),
     !.
+term_html(Undef, _Vars, WriteOptions) -->
+    { nonvar(Undef),
+      Undef = '$wfs_undefined'(Term),
+      !
+    },
+    html(span(class(wfs_undefined), \term(Term, WriteOptions))).
 term_html(Term, _Vars, WriteOptions) -->
     term(Term, WriteOptions).
 
@@ -704,6 +726,7 @@ prolog_help:show_html_hook(HTML) :-
 
 sandbox:safe_primitive(pengines_io:pengine_listing(_)).
 sandbox:safe_primitive(pengines_io:pengine_nl).
+sandbox:safe_primitive(pengines_io:pengine_tab(_)).
 sandbox:safe_primitive(pengines_io:pengine_flush_output).
 sandbox:safe_primitive(pengines_io:pengine_print(_)).
 sandbox:safe_primitive(pengines_io:pengine_write(_)).
@@ -734,6 +757,7 @@ sandbox:safe_meta(pengines_io:pengine_format(Format, Args), Calls) :-
 
 pengine_io_predicate(writeln(_)).
 pengine_io_predicate(nl).
+pengine_io_predicate(tab(_)).
 pengine_io_predicate(flush_output).
 pengine_io_predicate(format(_)).
 pengine_io_predicate(format(_,_)).
