@@ -68,7 +68,7 @@ extern "C" {
 /* PLVERSION_TAG: a string, normally "", but for example "rc1" */
 
 #ifndef PLVERSION
-#define PLVERSION 80502
+#define PLVERSION 80503
 #endif
 #ifndef PLVERSION_TAG
 #define PLVERSION_TAG ""
@@ -1158,7 +1158,8 @@ PL_EXPORT(int)	PL_current_prolog_flag(atom_t name, int type, void *ptr);
 #define PL_VERSION_VM		6	/* VM signature */
 #define PL_VERSION_BUILT_IN	7	/* Built-in predicate signature */
 
-PL_EXPORT(unsigned int) PL_version(int which);
+#define PL_version(id) PL_version_info(id)
+PL_EXPORT(unsigned int) PL_version_info(int which);
 
 
 		/********************************
@@ -1271,7 +1272,9 @@ typedef struct
 { int	(*unify)(term_t t, void *handle);	/* implementation --> Prolog */
   int   (*get)(term_t t, void **handle);	/* Prolog --> implementation */
   void	(*activate)(int active);		/* (de)activate */
-  intptr_t	magic;					/* PROFTYPE_MAGIC */
+  void  (*release)(void *handle);		/* Release handle */
+  void *dummy[4];				/* reserved */
+  intptr_t	magic;				/* PROFTYPE_MAGIC */
 } PL_prof_type_t;
 
 PL_EXPORT(int)		PL_register_profile_type(PL_prof_type_t *type);
