@@ -1578,6 +1578,7 @@ initPrologFlags(void)
 		PLFLAG_DOT_IN_ATOM);
   setPrologFlag("toplevel_var_size", FT_INTEGER, 1000);
   setPrologFlag("toplevel_print_anon", FT_BOOL, TRUE, 0);
+  setPrologFlag("toplevel_name_variables", FT_BOOL, TRUE, 0);
   setPrologFlag("toplevel_prompt", FT_ATOM, "~m~d~l~! ?- ");
   setPrologFlag("file_name_variables", FT_BOOL, FALSE, PLFLAG_FILEVARS);
   setPrologFlag("fileerrors", FT_BOOL, TRUE, PLFLAG_FILEERRORS);
@@ -1743,12 +1744,12 @@ abi_version_dict(term_t dict)
 			  ATOM_built_in };
   term_t values = PL_new_term_refs(6);
 
-  return ( PL_unify_integer(values+0, PL_version(PL_VERSION_FLI)) &&
-	   PL_unify_integer(values+1, PL_version(PL_VERSION_REC)) &&
-	   PL_unify_integer(values+2, PL_version(PL_VERSION_QLF)) &&
-	   PL_unify_integer(values+3, PL_version(PL_VERSION_QLF_LOAD)) &&
-	   PL_unify_integer(values+4, PL_version(PL_VERSION_VM)) &&
-	   PL_unify_integer(values+5, PL_version(PL_VERSION_BUILT_IN)) &&
+  return ( PL_unify_integer(values+0, PL_version_info(PL_VERSION_FLI)) &&
+	   PL_unify_integer(values+1, PL_version_info(PL_VERSION_REC)) &&
+	   PL_unify_integer(values+2, PL_version_info(PL_VERSION_QLF)) &&
+	   PL_unify_integer(values+3, PL_version_info(PL_VERSION_QLF_LOAD)) &&
+	   PL_unify_integer(values+4, PL_version_info(PL_VERSION_VM)) &&
+	   PL_unify_integer(values+5, PL_version_info(PL_VERSION_BUILT_IN)) &&
 
 	   PL_put_dict(dict, ATOM_abi, 6, keys, values) );
 }
