@@ -35,8 +35,6 @@
 
 #ifndef PL_INLINE_H_INCLUDED
 #define PL_INLINE_H_INCLUDED
-#undef LD
-#define LD LOCAL_LD
 
 #ifdef __WINDOWS__
   #include <windows.h>
@@ -266,14 +264,14 @@ fetchFunctorArray(size_t index)
   return GD->functors.array.blocks[idx][index];
 }
 
+#define pushVolatileAtom(a) LDFUNC(pushVolatileAtom, a)
 static inline void
-pushVolatileAtom__LD(atom_t a ARG_LD)
+pushVolatileAtom(DECL_LD atom_t a)
 { LD->atoms.unregistering = a;
   if ( GD->atoms.gc_active )
     markAtom(a);
 }
 
-#define pushVolatileAtom(a) pushVolatileAtom__LD(a PASS_LD)
 
 
 		 /*******************************
@@ -392,8 +390,9 @@ value need not be trailed.
 Note that the local stack is always _above_ the global stack.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+#define Trail(p, v) LDFUNC(Trail, p, v)
 static inline void
-Trail__LD(Word p, word v ARG_LD)
+Trail(DECL_LD Word p, word v)
 { DEBUG(CHK_SECURE, assert(tTop+1 <= tMax));
 
   if ( (void*)p >= (void*)lBase || p < LD->mark_bar )
@@ -402,10 +401,11 @@ Trail__LD(Word p, word v ARG_LD)
 }
 
 
-#define consPtrB(p, base, ts)	consPtr__LD(p, (uintptr_t)(base), (ts) PASS_LD)
+#define consPtrB(p, base, ts)	f_consPtr(p, (uintptr_t)(base), ts)
 #define consPtr(p, ts)		consPtrB(p, LD->bases[(ts)&STG_MASK], (ts))
+#define f_consPtr(p, base, ts) LDFUNC(f_consPtr, p, base, ts)
 static inline word
-consPtr__LD(void *p, uintptr_t base, word ts ARG_LD)
+f_consPtr(DECL_LD void *p, uintptr_t base, word ts)
 { uintptr_t v = (uintptr_t) p;
 
   v -= base;
@@ -414,9 +414,12 @@ consPtr__LD(void *p, uintptr_t base, word ts ARG_LD)
 }
 
 
-#if ALIGNOF_DOUBLE != ALIGNOF_VOIDP
+#if ALIGNOF_DOUBLE == ALIGNOF_VOIDP
+#define valFloat(w) (*(double *)valIndirectP(w))
+#else
+#define valFloat(w) LDFUNC(valFloat, w)
 static inline double
-valFloat__LD(word w ARG_LD)
+valFloat(DECL_LD word w)
 { Word p = valIndirectP(w);
   double d;
 
@@ -425,6 +428,41 @@ valFloat__LD(word w ARG_LD)
 }
 #endif
 
+
+#if ALIGNOF_INT64_T == ALIGNOF_VOIDP
+#define valBignum(w) (*(int64_t *)valIndirectP(w))
+#else
+#define valBignum(w) LDFUNC(valBignum, w)
+static inline int64_t
+valBignum(DECL_LD word w)
+{ Word p = valIndirectP(w);
+  union
+  { int64_t i;
+    word w[WORDS_PER_INT64];
+  } val;
+
+#if ( SIZEOF_VOIDP == 4 )
+  val.w[0] = p[0];
+  val.w[1] = p[1];
+#else
+#error "Unsupported int64_t alignment conversion"
+#endif
+
+  return val.i;
+}
+#endif
+
+
+#define valHandle(r) LDFUNC(valHandle, r)
+static inline word
+valHandle(DECL_LD term_t r)
+{ Word p = valTermRef(r);
+
+  deRef(p);
+  return *p;
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 linkValI(p)  is  an  inlined  version  of  linkVal()  that  assumes  the
 dereferenced value of `p` lives on  the   global  stack. It has no error
@@ -432,9 +470,9 @@ checking (unless compiled for debugging) and fetches the base address of
 the global stack only once.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define linkValI(p)		linkValI__LD(p PASS_LD)
+#define linkValI(p) LDFUNC(linkValI, p)
 static inline word
-linkValI__LD(Word p ARG_LD)
+linkValI(DECL_LD Word p)
 { word w = *p;
   uintptr_t gb = LD->bases[STG_GLOBAL];
 
@@ -455,13 +493,15 @@ linkValI__LD(Word p ARG_LD)
   }
 }
 
+#define is_signalled(_) LDFUNC(is_signalled, _)
 static inline int
-is_signalled(ARG1_LD)
+is_signalled(DECL_LD)
 { return HAS_LD && unlikely((LD->signal.pending[0]|LD->signal.pending[1]) != 0);
 }
 
+#define register_attvar(gp) LDFUNC(register_attvar, gp)
 static inline void
-register_attvar(Word gp ARG_LD)
+register_attvar(DECL_LD Word gp)
 { if ( LD->attvar.attvars )
   { *gp = makeRefG(LD->attvar.attvars);
     DEBUG(MSG_ATTVAR_LINK,
@@ -475,8 +515,9 @@ register_attvar(Word gp ARG_LD)
   LD->attvar.attvars = gp;
 }
 
+#define visibleClause(cl, gen) LDFUNC(visibleClause, cl, gen)
 static inline int
-visibleClause__LD(Clause cl, gen_t gen ARG_LD)
+visibleClause(DECL_LD Clause cl, gen_t gen)
 { gen_t c, e;
 
   c = cl->generation.created;
@@ -493,14 +534,15 @@ visibleClause__LD(Clause cl, gen_t gen ARG_LD)
 
   if ( unlikely(gen >= LD->transaction.gen_base) &&
        true(cl->predicate, P_DYNAMIC) )
-    return transaction_visible_clause(cl, gen PASS_LD);
+    return transaction_visible_clause(cl, gen);
 
   return FALSE;
 }
 
+#define visibleClauseCNT(cl, gen) LDFUNC(visibleClauseCNT, cl, gen)
 static inline int
-visibleClauseCNT__LD(Clause cl, gen_t gen ARG_LD)
-{ if ( likely(visibleClause__LD(cl, gen PASS_LD)) )
+visibleClauseCNT(DECL_LD Clause cl, gen_t gen)
+{ if ( likely(visibleClause(cl, gen)) )
     return TRUE;
   LD->clauses.erased_skipped++;
   return FALSE;
@@ -511,8 +553,9 @@ global_generation(void)
 { return GD->_generation;
 }
 
+#define current_generation(def) LDFUNC(current_generation, def)
 static inline gen_t
-current_generation(Definition def ARG_LD)
+current_generation(DECL_LD Definition def)
 { if ( unlikely(!!LD->transaction.generation) && def && true(def, P_DYNAMIC) )
   { return LD->transaction.generation;
   } else
@@ -520,8 +563,9 @@ current_generation(Definition def ARG_LD)
   }
 }
 
+#define next_generation(def) LDFUNC(next_generation, def)
 static inline gen_t
-next_generation(Definition def ARG_LD)
+next_generation(DECL_LD Definition def)
 { if ( unlikely(!!LD->transaction.generation) && def && true(def, P_DYNAMIC) )
   { if ( LD->transaction.generation < LD->transaction.gen_max )
       return ++LD->transaction.generation;
@@ -537,8 +581,9 @@ next_generation(Definition def ARG_LD)
   }
 }
 
+#define max_generation(def) LDFUNC(max_generation, def)
 static inline gen_t
-max_generation(Definition def ARG_LD)
+max_generation(DECL_LD Definition def)
 { if ( unlikely(!!LD->transaction.generation) && def && true(def, P_DYNAMIC) )
     return LD->transaction.gen_max;
   else
@@ -556,8 +601,9 @@ global_generation()  and  storing  the  generation  in  our  frame,  our
 generation is updated and thus no harm is done.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+#define setGenerationFrame(fr) LDFUNC(setGenerationFrame, fr)
 static inline void
-setGenerationFrame__LD(LocalFrame fr ARG_LD)
+setGenerationFrame(DECL_LD LocalFrame fr)
 { if ( unlikely(LD->transaction.generation &&
 		true(fr->predicate, P_DYNAMIC)) )
   { setGenerationFrameVal(fr, LD->transaction.generation);
@@ -571,13 +617,52 @@ setGenerationFrame__LD(LocalFrame fr ARG_LD)
   }
 }
 
+		 /*******************************
+		 *	      INDEXING		*
+		 *******************************/
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+murmur_key() is used to quickly compute  a   key  from indirect data for
+clause indexing. ptr is the start of the  data. It is word aligned and n
+is the size in bytes, a multiple of sizeof(word). If the indirect is too
+long we has based on the start, end and length.
+
+The hash should not conflict  with   a  functor_t  (hence the STG_GLOBAL
+mask) and may never be 0.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static inline word
+murmur_key(void *ptr, size_t n)
+{ word k;
+
+  DEBUG(0, assert(n%sizeof(word) == 0));
+
+  if ( n > sizeof(word)*4 )
+  { word data[3];
+
+    data[0] = ((Word)ptr)[0];
+    data[1] = ((Word)ptr)[n/sizeof(word)-1];
+    data[2] = n;
+
+    k = MurmurHashAligned2(data, sizeof(word)*3, MURMUR_SEED);
+  } else
+  { k = MurmurHashAligned2(ptr, n, MURMUR_SEED);
+  }
+
+  k &= ~((word)STG_GLOBAL);
+  if ( !k ) k = 1;
+
+  return k;
+}
+
 		 /*******************************
 		 *	      THREADS		*
 		 *******************************/
 
 #ifdef O_PLMT
+#define acquire_ldata(info) LDFUNC(acquire_ldata, info)
 static inline PL_local_data_t *
-acquire_ldata__LD(PL_thread_info_t *info ARG_LD)
+acquire_ldata(DECL_LD PL_thread_info_t *info)
 { PL_local_data_t *ld = info->thread_data;
   LD->thread.info->access.ldata = ld;
   if ( ld && ld->magic == LD_MAGIC )
